from __future__ import annotations

import os
import typing as t
from abc import ABCMeta
from copy import deepcopy
from pathlib import Path
from collections import defaultdict

import yaml
from loguru import logger

from starwhale.utils import console, validate_obj_name, get_downloadable_sw_version
from starwhale.consts import (
    PythonRunEnv,
    SW_IMAGE_FMT,
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    SW_PYPI_PKG_NAME,
    DEFAULT_PAGE_SIZE,
    ENV_SW_IMAGE_REPO,
    DEFAULT_IMAGE_REPO,
    DEFAULT_MANIFEST_NAME,
    DEFAULT_PYTHON_VERSION,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import move_dir, ensure_dir, ensure_file, get_path_created_time
from starwhale.base.type import URIType, BundleType, InstanceType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.http import ignore_error
from starwhale.utils.venv import (
    detect_pip_req,
    venv_install_req,
    conda_install_req,
    create_python_env,
    restore_python_env,
    activate_python_env,
    dump_python_dep_env,
    DUMP_USER_PIP_REQ_FNAME,
)
from starwhale.base.bundle import BaseBundle, LocalStorageBundleMixin
from starwhale.utils.error import (
    ExistedError,
    NotFoundError,
    NoSupportError,
    ConfigFormatError,
)
from starwhale.utils.progress import run_with_progress_bar
from starwhale.base.bundle_copy import BundleCopy

from .store import RuntimeStorage


class RuntimeConfig(object):
    def __init__(
        self,
        name: str,
        mode: str = PythonRunEnv.VENV,
        python_version: str = DEFAULT_PYTHON_VERSION,
        pip_req: str = DUMP_USER_PIP_REQ_FNAME,
        **kw: t.Any,
    ) -> None:
        self.name = name.strip().lower()
        self.mode = mode
        self.python_version = python_version.strip()
        self.pip_req = pip_req
        self.kw = kw
        self.starwhale_version = get_downloadable_sw_version()

        self._do_validate()

    def _do_validate(self) -> None:
        ok, reason = validate_obj_name(self.name)
        if not ok:
            raise ConfigFormatError(f"name:{self.name}, reason:{reason}")

        if self.mode not in (PythonRunEnv.CONDA, PythonRunEnv.VENV):
            raise ConfigFormatError(f"{self.mode} no support")

        if not self.python_version.startswith("3."):
            raise ConfigFormatError(f"only support Python3, set {self.python_version}")

        # TODO: add more validators

    @classmethod
    def create_by_yaml(cls, path: Path) -> RuntimeConfig:
        c = yaml.safe_load(path.open())
        return cls(**c)

    def as_dict(self) -> t.Dict[str, t.Any]:
        _d = deepcopy(self.__dict__)
        _d.pop("kw", None)
        return _d


class Runtime(BaseBundle):
    __metaclass__ = ABCMeta

    @classmethod
    def restore(cls, workdir: Path) -> None:
        StandaloneRuntime.restore(workdir)

    @classmethod
    def create(
        cls,
        workdir: t.Union[str, Path],
        name: str,
        python_version: str = DEFAULT_PYTHON_VERSION,
        mode: str = PythonRunEnv.VENV,
        force: bool = False,
    ) -> None:
        StandaloneRuntime.create(
            workdir=workdir,
            name=name,
            python_version=python_version,
            mode=mode,
            force=force,
        )

    @classmethod
    def get_runtime(cls, uri: URI) -> Runtime:
        _cls = cls._get_cls(uri)
        return _cls(uri)

    @classmethod
    def _get_cls(  # type: ignore
        cls,
        uri: URI,
    ) -> t.Union[t.Type[StandaloneRuntime], t.Type[CloudRuntime]]:
        if uri.instance_type == InstanceType.STANDALONE:
            return StandaloneRuntime
        elif uri.instance_type == InstanceType.CLOUD:
            return CloudRuntime
        else:
            raise NoSupportError(f"runtime uri:{uri}")

    def __str__(self) -> str:
        return f"Starwhale Runtime: {self.uri}"

    @classmethod
    def copy(cls, src_uri: str, dest_uri: str, force: bool = False) -> None:
        bc = BundleCopy(src_uri, dest_uri, URIType.RUNTIME, force)
        bc.do()

    @classmethod
    def activate(cls, workdir: str, yaml_name: str) -> None:
        StandaloneRuntime.activate(workdir, yaml_name)


class StandaloneRuntime(Runtime, LocalStorageBundleMixin):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.STANDALONE
        self.store = RuntimeStorage(uri)
        self.tag = StandaloneTag(uri)
        self._manifest: t.Dict[str, t.Any] = {}  # TODO: use manifest classget_conda_env

    def info(self) -> t.Dict[str, t.Any]:
        return self._get_bundle_info()

    def add_tags(self, tags: t.List[str], quiet: bool = False) -> None:
        self.tag.add(tags, quiet)

    def remove_tags(self, tags: t.List[str], quiet: bool = False) -> None:
        self.tag.remove(tags, quiet)

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        _ok, _reason = move_dir(self.store.loc, self.store.recover_loc, force)
        _ok2, _reason2 = True, ""
        if self.store.snapshot_workdir.exists():
            _ok2, _reason2 = move_dir(
                self.store.snapshot_workdir, self.store.recover_snapshot_workdir, force
            )
        return _ok and _ok2, _reason + _reason2

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support short version to recover, today only support full-version
        dest_path = (
            self.store.bundle_dir / f"{self.uri.object.version}{BundleType.RUNTIME}"
        )
        _ok, _reason = move_dir(self.store.recover_loc, dest_path, force)
        _ok2, _reason2 = True, ""
        if self.store.recover_snapshot_workdir.exists():
            _ok2, _reason2 = move_dir(
                self.store.recover_snapshot_workdir, self.store.snapshot_workdir, force
            )
        return _ok and _ok2, _reason + _reason2

    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:

        # TODO: time order
        _r = []
        for _bf in self.store.iter_bundle_history():
            _r.append(
                dict(
                    version=_bf.version,
                    path=str(_bf.path.resolve()),
                    created_at=get_path_created_time(_bf.path),
                    size=_bf.path.stat().st_size,
                    tags=_bf.tags,
                )
            )
        return _r, {}

    def build(
        self,
        workdir: Path,
        yaml_name: str = DefaultYAMLName.RUNTIME,
        **kw: t.Any,
    ) -> None:
        # TODO: tune for no runtime.yaml file
        _mp = workdir / yaml_name
        _swrt_config = self._load_runtime_config(_mp)
        _pip_req_path = detect_pip_req(workdir, _swrt_config.pip_req)
        _python_version = _swrt_config.python_version

        operations = [
            (self._gen_version, 5, "gen version"),
            (self._prepare_snapshot, 5, "prepare snapshot"),
            (
                self._dump_dep,
                50,
                "dump python dependency",
                dict(
                    gen_all_bundles=kw.get("gen_all_bundles", False),
                    pip_req_path=_pip_req_path,
                    python_version=_python_version,
                    mode=_swrt_config.mode,
                    include_editable=kw.get("include_editable", False),
                    identity=_swrt_config.name,
                ),
            ),
            (
                self._dump_base_image,
                5,
                "dump base image",
                dict(config=_swrt_config),
            ),
            (
                self._render_manifest,
                5,
                "render manifest",
                dict(user_raw_config=_swrt_config.as_dict()),
            ),
            (self._make_tar, 20, "make runtime bundle", dict(ftype=BundleType.RUNTIME)),
            (self._make_latest_tag, 5, "make latest tag"),
        ]
        run_with_progress_bar("runtime bundle building...", operations)

    def _dump_base_image(self, config: RuntimeConfig) -> None:
        _repo = os.environ.get(ENV_SW_IMAGE_REPO, DEFAULT_IMAGE_REPO)
        _tag = config.starwhale_version or "latest"
        base_image = SW_IMAGE_FMT.format(repo=_repo, tag=_tag)

        console.print(
            f":rainbow: runtime docker image: [red]{base_image}[/]  :rainbow:"
        )
        self._manifest["base_image"] = base_image

    def _dump_dep(
        self,
        gen_all_bundles: bool = False,
        pip_req_path: str = DUMP_USER_PIP_REQ_FNAME,
        python_version: str = DEFAULT_PYTHON_VERSION,
        mode: str = PythonRunEnv.AUTO,
        include_editable: bool = False,
        identity: str = "",
    ) -> None:
        logger.info("[step:dep]start dump python dep...")
        _dep = dump_python_dep_env(
            dep_dir=self.store.snapshot_workdir / "dep",
            pip_req_fpath=pip_req_path,
            gen_all_bundles=gen_all_bundles,
            expected_runtime=python_version,
            mode=mode,
            include_editable=include_editable,
            identity=identity,
        )
        self._manifest["dep"] = _dep
        logger.info("[step:dep]finish dump dep")

    def _prepare_snapshot(self) -> None:
        logger.info("[step:prepare-snapshot]prepare runtime snapshot dirs...")

        # TODO: graceful clear?
        if self.store.snapshot_workdir.exists():
            raise ExistedError(str(self.store.snapshot_workdir))

        ensure_dir(self.store.snapshot_workdir)
        ensure_dir(self.store.venv_dir)
        ensure_dir(self.store.conda_dir)

        console.print(
            f":file_folder: workdir: [underline]{self.store.snapshot_workdir}[/]"
        )

    def _load_runtime_config(self, path: Path) -> RuntimeConfig:
        self._do_validate_yaml(path)
        return RuntimeConfig.create_by_yaml(path)

    def extract(self, force: bool = False, target: t.Union[str, Path] = "") -> Path:
        return self._do_extract(force, target)

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        rs = defaultdict(list)
        for _bf in RuntimeStorage.iter_all_bundles(
            project_uri, bundle_type=BundleType.RUNTIME, uri_type=URIType.RUNTIME
        ):
            # TODO: add more manifest info
            rs[_bf.name].append(
                {
                    "version": _bf.version,
                    "path": str(_bf.path.absolute()),
                    "created_at": get_path_created_time(_bf.path),
                    "size": _bf.path.stat().st_size,
                    "is_removed": _bf.is_removed,
                    "tags": _bf.tags,
                }
            )
        return rs, {}

    @classmethod
    def create(
        cls,
        workdir: t.Union[str, Path],
        name: str,
        python_version: str = DEFAULT_PYTHON_VERSION,
        mode: str = PythonRunEnv.VENV,
        force: bool = False,
    ) -> None:
        workdir = Path(workdir).absolute()
        config = RuntimeConfig(name=name, mode=mode, python_version=python_version)

        ensure_dir(workdir)
        _id = create_python_env(
            mode=mode,
            name=name,
            workdir=workdir,
            python_version=python_version,
            force=force,
        )

        _pkg_name = SW_PYPI_PKG_NAME
        if config.starwhale_version:
            _pkg_name = f"{_pkg_name}=={config.starwhale_version}"

        console.print(f":dog: install {_pkg_name} {mode}@{_id}...")
        if mode == PythonRunEnv.VENV:
            venv_install_req(_id, req=_pkg_name, enable_pre=True)
        elif mode == PythonRunEnv.CONDA:
            conda_install_req(env_name=_id, req=_pkg_name, enable_pre=True)

        cls.render_runtime_yaml(config, workdir, force)
        activate_python_env(mode=mode, identity=_id)

    @classmethod
    def activate(cls, workdir: str, yaml_name: str) -> None:
        _rf = Path(workdir) / yaml_name
        if not _rf.exists():
            raise NotFoundError(_rf)

        _run_config = RuntimeConfig.create_by_yaml(_rf)
        if _run_config.mode == PythonRunEnv.VENV:
            _id = os.path.join(workdir, "venv")
        else:
            _id = _run_config.name
        activate_python_env(mode=_run_config.mode, identity=_id)

    @staticmethod
    def render_runtime_yaml(
        config: RuntimeConfig, workdir: Path, force: bool = False
    ) -> None:
        _rm = workdir / DefaultYAMLName.RUNTIME

        if _rm.exists() and not force:
            raise ExistedError(f"{_rm} was already existed")

        ensure_dir(workdir)
        ensure_file(_rm, yaml.safe_dump(config.as_dict(), default_flow_style=False))

    @classmethod
    def restore(cls, workdir: Path) -> None:
        if not (
            workdir.exists()
            and (workdir / DEFAULT_MANIFEST_NAME).exists()
            and (workdir / "dep").exists()
        ):
            raise NoSupportError("only support swrt extract workdir")

        _manifest = yaml.safe_load((workdir / DEFAULT_MANIFEST_NAME).open())
        restore_python_env(
            workdir=workdir,
            mode=_manifest["dep"]["env"],
            python_version=_manifest["dep"]["python"],
            local_gen_env=_manifest["dep"]["local_gen_env"],
            pip_req=_manifest.get("user_raw_config", {}).get("pip_req", ""),
        )


class CloudRuntime(CloudRequestMixed, Runtime):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.CLOUD

    @classmethod
    @ignore_error(({}, {}))
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        crm = CloudRequestMixed()
        return crm._fetch_bundle_all_list(project_uri, URIType.RUNTIME, page, size)



import os
import typing as t
from pathlib import Path

from starwhale.utils import console, pretty_bytes
from starwhale.consts import DefaultYAMLName, DEFAULT_PAGE_IDX, DEFAULT_PAGE_SIZE
from starwhale.base.uri import URI
from starwhale.base.type import URIType, InstanceType
from starwhale.base.view import BaseTermView
from starwhale.core.dataset.store import DatasetStorage

from .model import Dataset


class DatasetTermView(BaseTermView):
    def __init__(self, dataset_uri: str) -> None:
        super().__init__()

        self.raw_uri = dataset_uri
        self.uri = URI(dataset_uri, expected_type=URIType.DATASET)
        self.dataset = Dataset.get_dataset(self.uri)

    @BaseTermView._simple_action_print
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.dataset.remove(force)

    @BaseTermView._simple_action_print
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.dataset.recover(force)

    @BaseTermView._pager
    @BaseTermView._header
    def history(
        self, fullname: bool = False
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        fullname = fullname or self.uri.instance_type == InstanceType.CLOUD
        return self._print_history(
            title="Dataset History List",
            history=self.dataset.history(),
            fullname=fullname,
        )

    @BaseTermView._header
    def info(self, fullname: bool = False) -> None:
        self._print_info(self.dataset.info(), fullname=fullname)

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _uri = URI(project_uri, expected_type=URIType.PROJECT)
        fullname = fullname or (_uri.instance_type == InstanceType.CLOUD)
        _datasets, _pager = Dataset.list(_uri, page, size)
        _data = BaseTermView.list_data(_datasets, show_removed, fullname)
        return _data, _pager

    @classmethod
    def build(
        cls, workdir: str, project: str, yaml_name: str = DefaultYAMLName.DATASET
    ) -> None:
        _dataset_uri = cls.prepare_build_bundle(
            workdir, project, yaml_name, URIType.DATASET
        )
        _ds = Dataset.get_dataset(_dataset_uri)
        _ds.build(Path(workdir), yaml_name)

    @classmethod
    def copy(cls, src_uri: str, dest_uri: str, force: bool = False) -> None:
        Dataset.copy(src_uri, dest_uri, force)
        console.print(":clap: copy done")

    @classmethod
    def render_fuse_json(cls, target: str, force: bool = False) -> None:
        if os.path.exists(target) and os.path.isdir(target):
            workdir = Path(target)
        else:
            uri = URI(target, URIType.DATASET)
            store = DatasetStorage(uri)
            workdir = store.loc

        console.print(f":crown: try to render fuse json@{workdir}...")
        Dataset.render_fuse_json(workdir, force)

    @BaseTermView._header
    def tag(self, tags: t.List[str], remove: bool = False, quiet: bool = False) -> None:
        if remove:
            console.print(f":golfer: remove tags {tags} @ {self.uri}...")
            self.dataset.remove_tags(tags, quiet)
        else:
            console.print(f":surfer: add tags {tags} @ {self.uri}...")
            self.dataset.add_tags(tags, quiet)


class DatasetTermViewRich(DatasetTermView):
    @classmethod
    @BaseTermView._pager
    @BaseTermView._header
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _datasets, _pager = super().list(
            project_uri, fullname, show_removed, page, size
        )
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": cls.place_holder_for_empty(),
        }

        cls.print_table("Dataset List", _datasets, custom_column=custom_column)
        return _datasets, _pager


class DatasetTermViewJson(DatasetTermView):
    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> None:
        _datasets, _pager = super().list(
            project_uri, fullname, show_removed, page, size
        )
        cls.pretty_json(_datasets)

    def info(self, fullname: bool = False) -> None:
        self.pretty_json(self.get_info_data(self.dataset.info(), fullname))


def get_term_view(ctx_obj: t.Dict) -> t.Type[DatasetTermView]:
    return (
        DatasetTermViewJson if ctx_obj.get("output") == "json" else DatasetTermViewRich
    )



import os
import typing as t

import click

from starwhale.consts import (
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    LOCAL_FUSE_JSON_NAME,
)
from starwhale.base.uri import URI
from starwhale.base.type import URIType, EvalTaskType
from starwhale.consts.env import SWEnv
from starwhale.core.job.view import JobTermView
from starwhale.core.dataset.store import DatasetStorage

from .view import get_term_view, ModelTermView


@click.group("model", help="Model management, build/copy/ppl/cmp/eval/extract...")
@click.pass_context
def model_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@model_cmd.command("build", help="[ONLY Standalone]Build starwhale model")
@click.argument("workdir", type=click.Path(exists=True, file_okay=False))
@click.option("-p", "--project", default="", help="Project URI")
@click.option(
    "-f",
    "--model-yaml",
    default=DefaultYAMLName.MODEL,
    help="mode yaml filename, default use ${workdir}/model.yaml file",
)
def _build(workdir: str, project: str, model_yaml: str) -> None:
    ModelTermView.build(workdir, project, model_yaml)


@model_cmd.command("tag", help="Model Tag Management, add or remove")
@click.argument("model")
@click.argument("tags", nargs=-1)
@click.option("-r", "--remove", is_flag=True, help="Remove tags")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Ignore tag name errors like name duplication, name absence",
)
def _tag(model: str, tags: t.List[str], remove: bool, quiet: bool) -> None:
    ModelTermView(model).tag(tags, remove, quiet)


@model_cmd.command("copy", help="Copy model, standalone <--> cloud")
@click.argument("src")
@click.argument("dest")
@click.option("-f", "--force", is_flag=True, help="Force to copy model")
def _copy(src: str, dest: str, force: bool) -> None:
    ModelTermView.copy(src, dest, force)


@model_cmd.command("info", help="Show model details")
@click.argument("model")
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _info(view: t.Type[ModelTermView], model: str, fullname: bool) -> None:
    view(model).info(fullname)


@model_cmd.command("list", help="List Model")
@click.option("--project", default="", help="Project URI")
@click.option("--fullname", is_flag=True, help="Show fullname of model version")
@click.option("--show-removed", is_flag=True, help="Show removed model")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for model list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for model list"
)
@click.pass_obj
def _list(
    view: t.Type[ModelTermView],
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
) -> None:
    view.list(project, fullname, show_removed, page, size)


@model_cmd.command("history", help="Show model history")
@click.argument("model")
@click.option("--fullname", is_flag=True, help="Show version fullname")
def _history(model: str, fullname: bool) -> None:
    ModelTermView(model).history(fullname)


@model_cmd.command("remove", help="Remove model")
@click.argument("model")
@click.option("-f", "--force", is_flag=True, help="Force to remove model")
def _remove(model: str, force: bool) -> None:
    click.confirm("continue to delete?", abort=True)
    ModelTermView(model).remove(force)


@model_cmd.command("recover", help="Recover model")
@click.argument("model")
@click.option("-f", "--force", is_flag=True, help="Force to recover model")
def _recover(model: str, force: bool) -> None:
    ModelTermView(model).recover(force)


@model_cmd.command(
    "extract", help="[ONLY Standalone]Extract local model bundle tar file into workdir"
)
@click.argument("model")
@click.option("-f", "--force", is_flag=True, help="Force to extract model bundle")
@click.option(
    "--target-dir",
    default="",
    help="Extract target dir.if omitted, swcli will use starwhale default workdir",
)
def _extract(model: str, force: bool, target_dir: str) -> None:
    ModelTermView(model).extract(force, target_dir)


@model_cmd.command("ppl")
@click.argument("target")
@click.option(
    "-f",
    "--model-yaml",
    default=DefaultYAMLName.MODEL,
    help="Model yaml filename, default use ${MODEL_DIR}/model.yaml file",
)
@click.option(
    "--status-dir",
    envvar=SWEnv.status_dir,
    default="/tmp/starwhale/ppl/status",
    help=f"PPL status dir, env is {SWEnv.status_dir}",
)
@click.option(
    "--log-dir",
    envvar=SWEnv.log_dir,
    default="/tmp/starwhale/ppl/log",
    help=f"PPL log dir, env is {SWEnv.log_dir}",
)
@click.option(
    "--result-dir",
    envvar=SWEnv.result_dir,
    default="/tmp/starwhale/ppl/result",
    help=f"PPL result dir, env is {SWEnv.result_dir}",
)
@click.option(
    "--input-config",
    envvar=SWEnv.input_config,
    default="/tmp/starwhale/ppl/config/input.json",
    help=f"Dataset input.json path or Dataset URI, which was generated by `swcli dataset render-fuse` command, env is {SWEnv.input_config}",
)
def _ppl(
    target: str,
    model_yaml: str,
    status_dir: str,
    log_dir: str,
    result_dir: str,
    input_config: str,
) -> None:
    """
    [ONLY Standalone]Run PPL

    TARGET: model uri or model workdir path, in Starwhale agent docker environment, only support workdir path.
    INPUT_JSON: Dataset uri or dataset/input.json path
    """
    if not os.path.exists(input_config):
        uri = URI(input_config, expected_type=URIType.DATASET)
        store = DatasetStorage(uri)
        input_config = str((store.snapshot_workdir / LOCAL_FUSE_JSON_NAME).absolute())

    # TODO: support render fuse json for cmp test

    ModelTermView.eval(
        target=target,
        yaml_name=model_yaml,
        typ=EvalTaskType.PPL,
        kw={
            "status_dir": status_dir,
            "log_dir": log_dir,
            "result_dir": result_dir,
            "input_config": input_config,
        },
    )


@model_cmd.command("cmp")
@click.argument("target")
@click.option(
    "-f",
    "--model-yaml",
    default=DefaultYAMLName.MODEL,
    help="Model yaml filename, default use ${MODEL_DIR}/model.yaml file",
)
@click.option(
    "--status-dir",
    envvar=SWEnv.status_dir,
    default="/tmp/starwhale/cmp/status",
    help=f"CMP status dir, env is {SWEnv.status_dir}",
)
@click.option(
    "--log-dir",
    envvar=SWEnv.log_dir,
    default="/tmp/starwhale/cmp/log",
    help=f"CMP log dir, env is {SWEnv.log_dir}",
)
@click.option(
    "--result-dir",
    envvar=SWEnv.result_dir,
    default="/tmp/starwhale/cmp/result",
    help=f"CMP result dir, env is {SWEnv.result_dir}",
)
@click.option(
    "--input-config",
    envvar=SWEnv.input_config,
    default="/tmp/starwhale/cmp/config/input.json",
    help=f"CMP input.json path, env is {SWEnv.input_config}",
)
def _cmp(
    target: str,
    model_yaml: str,
    status_dir: str,
    log_dir: str,
    result_dir: str,
    input_config: str,
) -> None:
    """
    [ONLY Standalone]Run CMP, compare inference output with label, then generate result jsonline file.

    TARGET: model uri or model workdir path, in Starwhale Agent Docker Environment, only support workdir path.
    """
    ModelTermView.eval(
        target=target,
        yaml_name=model_yaml,
        typ=EvalTaskType.CMP,
        kw={
            "status_dir": status_dir,
            "log_dir": log_dir,
            "result_dir": result_dir,
            "input_config": input_config,
        },
    )


@model_cmd.command("eval")
@click.argument("model")
@click.option(
    "--dataset",
    required=True,
    multiple=True,
    help="Dataset URI, one or more",
)
@click.option("--name", help="Job name")
@click.option("--desc", help="Job description")
@click.option("-p", "--project", default="", help="Project URI")
def _eval(model: str, dataset: t.List[str], name: str, desc: str, project: str) -> None:
    """
    [ONLY Standalone]Create as new job for model evaluation

    MODEL: model uri or model workdir path
    """
    JobTermView.create(
        project_uri=project,
        model_uri=model,
        dataset_uris=dataset,
        runtime_uri="",
        name=name,
        desc=desc,
        use_docker=False,
        gencmd=False,
        phase=EvalTaskType.ALL,
    )



import typing as t
from pathlib import Path

from starwhale.consts import DEFAULT_MANIFEST_NAME
from starwhale.base.type import URIType, BundleType
from starwhale.base.store import BaseStorage


class ModelStorage(BaseStorage):
    def _guess(self) -> t.Tuple[Path, str]:
        return self._guess_for_bundle()

    @property
    def bundle_type(self) -> str:
        return BundleType.MODEL

    @property
    def uri_type(self) -> str:
        return URIType.MODEL

    @property
    def src_dir(self):
        return self.snapshot_workdir / "src"

    @property
    def recover_loc(self) -> Path:
        return self._get_recover_loc_for_bundle()

    @property
    def manifest_path(self) -> Path:
        return self.snapshot_workdir / DEFAULT_MANIFEST_NAME

    @property
    def snapshot_workdir(self) -> Path:
        return self._get_snapshot_workdir_for_bundle()

    @property
    def recover_snapshot_workdir(self) -> Path:
        return self._get_recover_snapshot_workdir_for_bundle()



import os
import re
import time
import uuid
import base64
import random
import string
import typing as t
import hashlib
import platform
from datetime import datetime

from rich.console import Console

from starwhale import __version__
from starwhale.consts import FMT_DATETIME, SW_DEV_DUMMY_VERSION

console = Console(soft_wrap=True)
now_str = lambda: datetime.now().astimezone().strftime(FMT_DATETIME)


def timestamp_to_datatimestr(timestamp: float) -> str:
    ts = time.localtime(timestamp)
    return time.strftime(FMT_DATETIME, ts)


def gen_uniq_version(feature: str = "") -> str:
    # version = ${timestamp:8} + ${feature:8} + ${randstr:4}
    timestamp = "".join(str(uuid.uuid1()).split("-")[0])
    feature = hashlib.sha256((feature or random_str()).encode()).hexdigest()[:7]
    randstr = random_str(cnt=4)
    bstr = base64.b32encode(f"{timestamp}{feature}{randstr}".encode()).decode("ascii")
    # TODO: add test for uniq and number
    return bstr.lower().strip("=")


def random_str(cnt: int = 8) -> str:
    return "".join(random.sample(string.ascii_lowercase + string.digits, cnt))


def in_dev() -> bool:
    return not in_production()


def in_production() -> bool:
    return os.environ.get("SW_PRODUCTION", "") == "1"


def is_windows() -> bool:
    # TODO: for windows nt?
    return platform.system() == "Windows"


def is_darwin(arm: bool = False) -> bool:
    if platform.system() != "Darwin":
        return False
    return not arm or platform.processor() == "arm"


def is_linux() -> bool:
    return platform.system() in ("Linux", "Unix")


def fmt_http_server(server: str, https: bool = False) -> str:
    server = server.strip().strip("/")
    if not server:
        raise Exception("no server addr")

    if server.startswith(("http://", "https://")):
        return server
    else:
        prefix = "https" if https else "http"
        return f"{prefix}://{server}"


_bytes_map = {
    "k": 1024,
    "kb": 1024,
    "m": 1024 * 1024,
    "mb": 1024 * 1024,
    "g": 1024 * 1024 * 1024,
    "gb": 1024 * 1024 * 1024,
}


def convert_to_bytes(s: t.Union[str, int]) -> int:
    if isinstance(s, int):
        return s

    s = s.strip().lower()
    for f in ("k", "m", "g"):
        if s.endswith((f, f"{f}b")):
            return _bytes_map[f] * int(s.split(f)[0])
    else:
        return int(s)


_bytes_progress = ("B", "KB", "MB", "GB", "TB", "PB")


def pretty_bytes(b: t.Union[int, float]) -> str:
    def _c(b: t.Union[int, float], idx: int) -> str:
        if b < 1024 or (idx + 1 == len(_bytes_progress)):
            return f"{b:.2f}{_bytes_progress[idx]}"
        else:
            return _c(b / 1024, idx + 1)

    return _c(b, 0)


_valid_name_re = re.compile("^[a-zA-Z_][a-zA-Z0-9_.-]*$")


def validate_obj_name(name: str) -> t.Tuple[bool, str]:
    if len(name) < 1 or len(name) > 80:
        return (
            False,
            f"length should be between 1 and 80, but {name} has {len(name)} characters",
        )

    if not _valid_name_re.match(name):
        return (
            False,
            f"A name MUST only consist of letters A-Z a-z, digits 0-9, the hyphen character -, the underscore character _ and the dot character ., current name:{name}",
        )

    return True, ""


def get_downloadable_sw_version() -> str:
    _v = __version__
    return "" if _v == SW_DEV_DUMMY_VERSION else _v


def snake_to_camel(snake: str) -> str:
    parts = snake.split("_")
    return "".join(i.title() for i in parts)



import os
import typing as t
from pathlib import Path

from starwhale.utils import console, pretty_bytes, in_production
from starwhale.consts import DefaultYAMLName, DEFAULT_PAGE_IDX, DEFAULT_PAGE_SIZE
from starwhale.base.uri import URI
from starwhale.base.type import URIType, EvalTaskType, InstanceType
from starwhale.base.view import BaseTermView
from starwhale.core.model.store import ModelStorage

from .model import Model, StandaloneModel


class ModelTermView(BaseTermView):
    def __init__(self, model_uri: str) -> None:
        super().__init__()

        self.raw_uri = model_uri
        self.uri = URI(model_uri, expected_type=URIType.MODEL)
        self.model = Model.get_model(self.uri)

    @BaseTermView._simple_action_print
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.model.remove(force)

    @BaseTermView._simple_action_print
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.model.recover(force)

    @BaseTermView._header
    def info(self, fullname: bool = False) -> None:
        self._print_info(self.model.info(), fullname=fullname)

    @BaseTermView._pager
    @BaseTermView._header
    def history(
        self, fullname: bool = False
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        fullname = fullname or self.uri.instance_type == InstanceType.CLOUD
        return self._print_history(
            title="Model History List", history=self.model.history(), fullname=fullname
        )

    def extract(self, force: bool = False, target_dir: str = "") -> None:
        console.print(":oncoming_police_car: try to extract ...")
        path = self.model.extract(force, target_dir)
        console.print(f":clap: extracted @ {path.resolve()} :tada:")

    @classmethod
    def eval(
        cls,
        target: str,
        yaml_name: str = DefaultYAMLName.MODEL,
        typ: str = "",
        kw: t.Dict[str, t.Any] = {},
    ) -> None:
        if in_production() or (os.path.exists(target) and os.path.isdir(target)):
            workdir = Path(target)
        else:
            uri = URI(target, URIType.MODEL)
            store = ModelStorage(uri)
            workdir = store.loc

        if typ in (EvalTaskType.CMP, EvalTaskType.PPL):
            console.print(f":golfer: try to eval {typ} @ {workdir}...")
            StandaloneModel.eval_user_handler(
                typ,
                workdir,
                yaml_name=yaml_name,
                kw=kw,
            )
        else:
            pass

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _uri = URI(project_uri, expected_type=URIType.PROJECT)
        fullname = fullname or (_uri.instance_type == InstanceType.CLOUD)
        _models, _pager = Model.list(_uri, page, size)
        _data = BaseTermView.list_data(_models, show_removed, fullname)
        return _data, _pager

    @classmethod
    def build(
        cls, workdir: str, project: str, yaml_name: str = DefaultYAMLName.MODEL
    ) -> None:
        _model_uri = cls.prepare_build_bundle(
            workdir, project, yaml_name, URIType.MODEL
        )
        _m = Model.get_model(_model_uri)
        _m.build(Path(workdir), yaml_name)

    @classmethod
    def copy(cls, src_uri: str, dest_uri: str, force: bool = False) -> None:
        Model.copy(src_uri, dest_uri, force)
        console.print(":clap: copy done.")

    @BaseTermView._header
    def tag(self, tags: t.List[str], remove: bool = False, quiet: bool = False) -> None:
        if remove:
            console.print(f":golfer: remove tags [red]{tags}[/] @ {self.uri}...")
            self.model.remove_tags(tags, quiet)
        else:
            console.print(f":surfer: add tags [red]{tags}[/] @ {self.uri}...")
            self.model.add_tags(tags, quiet)


class ModelTermViewRich(ModelTermView):
    @classmethod
    @BaseTermView._pager
    @BaseTermView._header
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _models, _pager = super().list(project_uri, fullname, show_removed, page, size)
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": cls.place_holder_for_empty(""),
        }

        cls.print_table("Model List", _models, custom_column=custom_column)
        return _models, _pager


class ModelTermViewJson(ModelTermView):
    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> None:
        _models, _pager = super().list(project_uri, fullname, show_removed, page, size)
        cls.pretty_json(_models)

    def info(self, fullname: bool = False) -> None:
        self.pretty_json(self.get_info_data(self.model.info(), fullname))


def get_term_view(ctx_obj: t.Dict) -> t.Type[ModelTermView]:
    return ModelTermViewJson if ctx_obj.get("output") == "json" else ModelTermViewRich



from __future__ import annotations

import os
import typing as t
from abc import ABCMeta
from copy import deepcopy
from pathlib import Path
from collections import defaultdict

import yaml
from fs import open_fs
from loguru import logger
from fs.copy import copy_fs, copy_file

from starwhale.utils import console
from starwhale.consts import (
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    DEFAULT_COPY_WORKERS,
    DEFAULT_STARWHALE_API_VERSION,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import move_dir, ensure_dir
from starwhale.base.type import URIType, BundleType, EvalTaskType, InstanceType
from starwhale.base.cloud import CloudRequestMixed, CloudBundleModelMixin
from starwhale.utils.http import ignore_error
from starwhale.utils.load import import_cls
from starwhale.base.bundle import BaseBundle, LocalStorageBundleMixin
from starwhale.utils.error import ExistedError, NoSupportError, FileFormatError
from starwhale.utils.progress import run_with_progress_bar
from starwhale.base.bundle_copy import BundleCopy

from .store import ModelStorage


class ModelRunConfig(object):

    # TODO: use attr to tune class
    def __init__(
        self,
        ppl: str,
        runtime: str = "",
        pkg_data: t.Union[t.List[str], None] = None,
        exclude_pkg_data: t.Union[t.List[str], None] = None,
        envs: t.Union[t.List[str], None] = None,
        **kw: t.Any,
    ):
        self.ppl = ppl.strip()
        self.runtime = runtime.strip()
        self.pkg_data = pkg_data or []
        self.exclude_pkg_data = exclude_pkg_data or []
        self.envs = envs or []
        self.kw = kw

        self._do_validate()

    def _do_validate(self) -> None:
        if not self.ppl:
            raise FileFormatError("need ppl field")

    def __str__(self) -> str:
        return f"Model Run Config: ppl -> {self.ppl}"

    def __repr__(self) -> str:
        return f"Model Run Config: ppl -> {self.ppl}, runtime -> {self.runtime}"

    def as_dict(self) -> t.Dict[str, t.Any]:
        return deepcopy(self.__dict__)


class ModelConfig(object):

    # TODO: use attr to tune class
    def __init__(
        self,
        name: str,
        model: t.List[str],
        run: t.Dict[str, t.Any],
        config: t.List[str] = [],
        desc: str = "",
        tag: t.List[str] = [],
        version: str = DEFAULT_STARWHALE_API_VERSION,
        **kw: t.Any,
    ):

        # TODO: format model name
        self.name = name
        self.model = model or []
        self.config = config or []
        # TODO: support artifacts: local or remote
        self.run = ModelRunConfig(**run)
        self.desc = desc
        self.tag = tag
        self.version = version
        self.kw = kw

        self._do_validate()

    def _do_validate(self) -> None:
        # TODO: use attr validator
        if not self.model:
            raise FileFormatError("need at least one model")

        # TODO: add more validation
        # TODO: add name check

    @classmethod
    def create_by_yaml(cls, path: Path) -> ModelConfig:
        c = yaml.safe_load(path.open())
        return cls(**c)

    def __str__(self) -> str:
        return f"Model Config: {self.name}"

    def __repr__(self) -> str:
        return f"Model Config: name -> {self.name}, model-> {self.model}"

    def as_dict(self) -> t.Dict[str, t.Any]:
        _r = deepcopy(self.__dict__)
        _r["run"] = self.run.as_dict()
        return _r


class Model(BaseBundle):
    __metaclass__ = ABCMeta

    def __str__(self) -> str:
        return f"Starwhale Model: {self.uri}"

    def eval(self) -> None:
        pass

    def ppl(
        self, workdir: Path, yaml_name: str = DefaultYAMLName.MODEL, **kw: t.Any
    ) -> None:
        pass

    def cmp(
        self, workdir: Path, yaml_name: str = DefaultYAMLName.MODEL, **kw: t.Any
    ) -> None:
        pass

    @classmethod
    def get_model(cls, uri: URI) -> Model:
        _cls = cls._get_cls(uri)
        return _cls(uri)

    @classmethod
    def _get_cls(cls, uri: URI) -> t.Union[t.Type[StandaloneModel], t.Type[CloudModel]]:  # type: ignore
        if uri.instance_type == InstanceType.STANDALONE:
            return StandaloneModel
        elif uri.instance_type == InstanceType.CLOUD:
            return CloudModel
        else:
            raise NoSupportError(f"model uri:{uri}")

    @classmethod
    def copy(cls, src_uri: str, dest_uri: str, force: bool = False) -> None:
        bc = BundleCopy(src_uri, dest_uri, URIType.MODEL, force)
        bc.do()


class StandaloneModel(Model, LocalStorageBundleMixin):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.STANDALONE
        self.store = ModelStorage(uri)
        self.tag = StandaloneTag(uri)
        self._manifest: t.Dict[str, t.Any] = {}  # TODO: use manifest classget_conda_env

    def add_tags(self, tags: t.List[str], quiet: bool = False) -> None:
        self.tag.add(tags, quiet)

    def remove_tags(self, tags: t.List[str], quiet: bool = False) -> None:
        self.tag.remove(tags, quiet)

    @classmethod
    def eval_user_handler(
        cls,
        typ: str,
        workdir: Path,
        yaml_name: str = DefaultYAMLName.MODEL,
        kw: t.Dict[str, t.Any] = {},
    ) -> None:
        from starwhale.api._impl.model import _RunConfig

        _mp = workdir / yaml_name
        _model_config = cls._load_model_config(_mp)
        _handler = _model_config.run.ppl

        _RunConfig.set_env(kw)
        console.print(f"try to import {_handler}@{workdir}...")
        _cls = import_cls(workdir, _handler)

        with _cls() as _obj:
            if typ == EvalTaskType.CMP:
                _obj._starwhale_internal_run_cmp()
            else:
                _obj._starwhale_internal_run_ppl()

        console.print(f":clap: finish run {typ}: {_obj}")

    def info(self) -> t.Dict[str, t.Any]:
        return self._get_bundle_info()

    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:

        _r = []
        for _bf in self.store.iter_bundle_history():
            _manifest = ModelStorage.get_manifest_by_path(
                _bf.path, BundleType.MODEL, URIType.MODEL
            )

            _r.append(
                dict(
                    name=self.name,
                    version=_bf.version,
                    path=str(_bf.path.resolve()),
                    tags=_bf.tags,
                    created_at=_manifest["created_at"],
                    size=_bf.path.stat().st_size,
                    runtime=_manifest["user_raw_config"]
                    .get("run", {})
                    .get("runtime", "--"),
                )
            )
        return _r, {}

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        _ok, _reason = move_dir(self.store.loc, self.store.recover_loc, force)
        _ok2, _reason2 = True, ""
        if self.store.snapshot_workdir.exists():
            _ok2, _reason2 = move_dir(
                self.store.snapshot_workdir, self.store.recover_snapshot_workdir, force
            )
        return _ok and _ok2, _reason + _reason2

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support short version to recover, today only support full-version
        dest_path = (
            self.store.bundle_dir / f"{self.uri.object.version}{BundleType.MODEL}"
        )
        _ok, _reason = move_dir(self.store.recover_loc, dest_path, force)
        _ok2, _reason2 = True, ""
        if self.store.recover_snapshot_workdir.exists():
            _ok2, _reason2 = move_dir(
                self.store.recover_snapshot_workdir, self.store.snapshot_workdir, force
            )
        return _ok and _ok2, _reason + _reason2

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        rs = defaultdict(list)
        for _bf in ModelStorage.iter_all_bundles(
            project_uri,
            bundle_type=BundleType.MODEL,
            uri_type=URIType.MODEL,
        ):
            _manifest = ModelStorage.get_manifest_by_path(
                _bf.path, BundleType.MODEL, URIType.MODEL
            )

            rs[_bf.name].append(
                {
                    "name": _manifest["name"],
                    "version": _bf.version,
                    "path": str(_bf.path.absolute()),
                    "size": _bf.path.stat().st_size,
                    "is_removed": _bf.is_removed,
                    "runtime": _manifest["user_raw_config"]
                    .get("run", {})
                    .get("runtime", "--"),
                    "created_at": _manifest["created_at"],
                    "tags": _bf.tags,
                }
            )
        return rs, {}

    def build(
        self, workdir: Path, yaml_name: str = DefaultYAMLName.MODEL, **kw: t.Any
    ) -> None:
        _mp = workdir / yaml_name
        _model_config = self._load_model_config(_mp)

        operations = [
            (self._gen_version, 5, "gen version"),
            (self._prepare_snapshot, 5, "prepare snapshot"),
            (
                self._copy_src,
                15,
                "copy src",
                dict(workdir=workdir, yaml_name=yaml_name, model_config=_model_config),
            ),
            (
                self._render_manifest,
                5,
                "render manifest",
                dict(user_raw_config=_model_config.as_dict()),
            ),
            (self._make_tar, 20, "build model bundle", dict(ftype=BundleType.MODEL)),
            (self._make_latest_tag, 5, "make latest tag"),
        ]
        run_with_progress_bar("model bundle building...", operations)

    @classmethod
    def _load_model_config(cls, yaml_path: Path) -> ModelConfig:
        cls._do_validate_yaml(yaml_path)
        _config = ModelConfig.create_by_yaml(yaml_path)

        if not _config.model:
            raise FileFormatError("model yaml no model")

        for _fpath in _config.model + _config.config:
            if not (yaml_path.parent / _fpath).exists():
                raise FileFormatError(f"model - {_fpath} is not existed")

        # TODO: add more model.yaml section validation
        # TODO: add 'swcli model check' cmd

        cls._load_config_envs(_config)
        return _config

    def _prepare_snapshot(self) -> None:
        logger.info("[step:prepare-snapshot]prepare model snapshot dirs...")

        if self.store.snapshot_workdir.exists():
            raise ExistedError(str(self.store.snapshot_workdir))

        ensure_dir(self.store.snapshot_workdir)
        ensure_dir(self.store.src_dir)

        # TODO: cleanup garbage dir
        # TODO: add lock/flag file for gc

        console.print(
            f":file_folder: workdir: [underline]{self.store.snapshot_workdir}[/]"
        )

    def _copy_src(
        self, workdir: Path, yaml_name: str, model_config: ModelConfig
    ) -> None:
        logger.info(
            f"[step:copy]start to copy src {workdir} -> {self.store.src_dir} ..."
        )
        console.print(":thumbs_up: try to copy source code files...")
        _mc = model_config

        workdir_fs = open_fs(str(workdir.resolve()))
        snapshot_fs = open_fs(str(self.store.snapshot_workdir.resolve()))
        src_fs = open_fs(str(self.store.src_dir.resolve()))
        # TODO: support glob pkg_data
        copy_file(workdir_fs, yaml_name, snapshot_fs, DefaultYAMLName.MODEL)
        copy_fs(
            workdir_fs,
            src_fs,
            walker=self._get_src_walker(
                workdir, _mc.run.pkg_data, _mc.run.exclude_pkg_data
            ),
            workers=DEFAULT_COPY_WORKERS,
        )

        for _fname in _mc.config + _mc.model:
            copy_file(workdir_fs, _fname, src_fs, _fname)

        logger.info("[step:copy]finish copy files")

    @classmethod
    def _load_config_envs(cls, _config: ModelConfig) -> None:
        for _env in _config.run.envs:
            _env = _env.strip()
            if not _env:
                continue
            _t = _env.split("=", 1)
            _k, _v = _t[0], "".join(_t[1:])

            if _k not in os.environ:
                os.environ[_k] = _v

    def extract(self, force: bool = False, target: t.Union[str, Path] = "") -> Path:
        return self._do_extract(force, target)


class CloudModel(CloudBundleModelMixin, Model):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.CLOUD

    @classmethod
    @ignore_error(({}, {}))
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        crm = CloudRequestMixed()
        return crm._fetch_bundle_all_list(project_uri, URIType.MODEL, page, size)



import os
import typing as t
import getpass
from pathlib import Path

import yaml

from starwhale.consts import (
    UserRoleType,
    SW_CLI_CONFIG,
    DEFAULT_PROJECT,
    DEFAULT_INSTANCE,
    SW_LOCAL_STORAGE,
    ENV_SW_CLI_CONFIG,
    STANDALONE_INSTANCE,
    LOCAL_CONFIG_VERSION,
)
from starwhale.utils.error import NotFoundError

from . import console, now_str, fmt_http_server
from .fs import ensure_dir, ensure_file

_config: t.Dict[str, t.Any] = {}
_CURRENT_SHELL_USERNAME = getpass.getuser()


def load_swcli_config() -> t.Dict[str, t.Any]:
    global _config

    if _config:
        ensure_dir(Path(_config["storage"]["root"]) / DEFAULT_PROJECT, recursion=True)
        return _config

    # TODO: add set_global_env func in cli startup
    fpath = get_swcli_config_path()

    if not os.path.exists(fpath):
        _config = render_default_swcli_config(fpath)
    else:
        with open(fpath) as f:
            _config = yaml.safe_load(f)

            _version = _config.get("version")
            if _version != LOCAL_CONFIG_VERSION:
                console.print(
                    f":cherries: {fpath} use unexpected version({_version}), swcli only support {LOCAL_CONFIG_VERSION} version."
                )
                console.print(
                    f":carrot: {fpath} will be upgraded to {LOCAL_CONFIG_VERSION} automatically."
                )
                _config = render_default_swcli_config(fpath)

    ensure_dir(Path(_config["storage"]["root"]) / DEFAULT_PROJECT, recursion=True)
    return _config


def render_default_swcli_config(fpath: str) -> t.Dict[str, t.Any]:
    from starwhale.base.type import InstanceType

    c = dict(
        version=LOCAL_CONFIG_VERSION,
        instances={
            STANDALONE_INSTANCE: dict(
                uri=DEFAULT_INSTANCE,
                user_name=_CURRENT_SHELL_USERNAME,
                current_project=DEFAULT_PROJECT,
                type=InstanceType.STANDALONE,
                updated_at=now_str(),  # type: ignore
            )
        },
        current_instance=DEFAULT_INSTANCE,
        storage=dict(root=str(SW_LOCAL_STORAGE.resolve())),
    )
    render_swcli_config(c, fpath)
    return c


def update_swcli_config(**kw: t.Any) -> None:
    c = load_swcli_config()
    # TODO: tune update config
    # TODO: add deepcopy for dict?
    c.update(kw)
    render_swcli_config(c)


def get_swcli_config_path() -> str:
    fpath = os.environ.get(ENV_SW_CLI_CONFIG, "")
    if not fpath or not os.path.exists(fpath):
        fpath = str(SW_CLI_CONFIG)
    return fpath


def render_swcli_config(c: t.Dict[str, t.Any], path: str = "") -> None:
    fpath = path or get_swcli_config_path()
    ensure_dir(os.path.dirname(fpath), recursion=True)
    ensure_file(fpath, yaml.safe_dump(c, default_flow_style=False), mode=0o600)


# TODO: abstract better common base or mixed class
class SWCliConfigMixed(object):
    def __init__(self, swcli_config: t.Union[t.Dict[str, t.Any], None] = None) -> None:
        self._config = swcli_config or load_swcli_config()

    @property
    def rootdir(self) -> Path:
        return Path(self._config["storage"]["root"])

    @property
    def workdir(self) -> Path:
        return self.rootdir / "workdir"

    @property
    def pkgdir(self) -> Path:
        return self.rootdir / "pkg"

    @property
    def dataset_dir(self) -> Path:
        return self.rootdir / "dataset"

    @property
    def eval_run_dir(self) -> Path:
        return self.rootdir / "run" / "eval"

    @property
    def sw_remote_addr(self) -> str:
        addr = self._current_instance_obj.get("uri", "")
        return fmt_http_server(addr)

    @property
    def user_name(self) -> str:
        return self._current_instance_obj.get("user_name", "")

    @property
    def _sw_token(self) -> str:
        return self._current_instance_obj.get("sw_token", "")

    @property
    def _current_instance_obj(self) -> t.Dict[str, t.Any]:
        return self._config.get("instances", {}).get(self.current_instance, {})

    @property
    def user_role(self) -> str:
        return self._current_instance_obj.get("user_role", "")

    @property
    def current_instance(self) -> str:
        return self._config["current_instance"]  # type: ignore

    def get_sw_instance_config(self, instance: str) -> t.Dict[str, t.Any]:
        instance = self._get_instance_alias(instance)
        return self._config["instances"].get(instance, {})

    def get_sw_token(self, instance: str) -> str:
        return self.get_sw_instance_config(instance).get("sw_token", "")

    def _get_instance_alias(self, instance: str) -> str:
        if not instance:
            return self.current_instance

        if instance not in self._config["instances"]:
            for k, v in self._config["instances"].items():
                if v["uri"] == instance:
                    return k

        return instance

    @property
    def current_project(self) -> str:
        return self._current_instance_obj.get("current_project", "")

    def select_current_default(self, instance: str, project: str = "") -> None:
        instance = self._get_instance_alias(instance)

        if instance not in self._config["instances"]:
            raise NotFoundError(f"need to login instance {instance}")

        self._config["current_instance"] = instance
        if project:
            if (
                instance == STANDALONE_INSTANCE
                and not (self.rootdir / project).exists()
            ):
                raise NotFoundError(f"need to create project {project}")
            # TODO: check cloud project existence
            self._config["instances"][instance]["current_project"] = project

        update_swcli_config(**self._config)

    def delete_instance(self, uri: str) -> None:
        if uri == STANDALONE_INSTANCE:
            return

        _insts = self._config["instances"]
        _alias = uri
        if uri in _insts:
            _insts.pop(uri)
        else:
            for k, v in _insts.items():
                if v.get("uri") == uri:
                    _insts.pop(k)
                    _alias = uri

        if _alias == self._config["current_instance"]:
            self._config["current_instance"] = DEFAULT_INSTANCE
        update_swcli_config(**self._config)

    def update_instance(
        self,
        uri: str,
        user_name: str = _CURRENT_SHELL_USERNAME,
        user_role: str = UserRoleType.NORMAL,
        sw_token: str = "",
        alias: str = "",
    ) -> None:
        from starwhale.base.type import InstanceType

        # TODO: abstrace instance class
        uri = uri.strip()
        if not uri.startswith(("http://", "https://")):
            uri = f"http://{uri}"

        alias = alias or uri
        if alias == STANDALONE_INSTANCE:
            console.print(f":person_running: skip {STANDALONE_INSTANCE} update")
            return

        # TODO: add more instance list and search
        _instances: t.Dict[str, t.Dict[str, str]] = self._config["instances"]
        if alias not in _instances:
            _instances[alias] = {}

        _instances[alias].update(
            uri=uri,
            user_name=user_name,
            user_role=user_role,
            sw_token=sw_token,
            type=InstanceType.CLOUD,
            updated_at=now_str(),  # type: ignore
        )

        update_swcli_config(**self._config)



import os
import sys
import json
import typing as t
from functools import wraps

import yaml
from rich import box
from rich import print as rprint
from rich.panel import Panel
from rich.table import Table
from rich.pretty import Pretty
from rich.console import RenderableType

from starwhale.utils import console, pretty_bytes, snake_to_camel
from starwhale.consts import UserRoleType, SHORT_VERSION_CNT
from starwhale.base.uri import URI
from starwhale.base.type import URIType
from starwhale.utils.error import FileFormatError
from starwhale.utils.config import SWCliConfigMixed


class BaseTermView(SWCliConfigMixed):
    @staticmethod
    def _pager(func):
        @wraps(func)
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> None:
            def _print(_r: t.Dict[str, t.Any]) -> None:
                p = Panel(
                    (
                        f"Counts: [green] {_r['current']}/{_r['total']} [/] :sheep: ,"
                        f"[red] {_r['remain']} [/] items does not show."
                    ),
                    title="Count Details",
                    title_align="left",
                )
                rprint(p)

            rt = func(*args, **kwargs)  # type: ignore
            if isinstance(rt, tuple) and len(rt) == 2 and "total" in rt[1]:
                _print(rt[1])

        return _wrapper

    @staticmethod
    def _header(func):
        @wraps(func)
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> None:
            sw = SWCliConfigMixed()

            def _print() -> None:
                grid = Table.grid(expand=True)
                grid.add_column(justify="center", ratio=1)
                grid.add_column(justify="right")
                grid.add_row(
                    f":star: {sw.current_instance} ({sw._current_instance_obj['uri']}) :whale:",  # type: ignore
                    f":clown_face:{sw._current_instance_obj['user_name']}@{sw._current_instance_obj.get('user_role', UserRoleType.NORMAL)}",  # type: ignore
                )
                p = Panel(grid, title="Starwhale Instance", title_align="left")
                rprint(p)

            _print()
            return func(*args, **kwargs)  # type: ignore

        return _wrapper

    @staticmethod
    def _simple_action_print(func):
        @wraps(func)
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> None:
            rt = func(*args, **kwargs)

            if isinstance(rt, tuple) and len(rt) == 2:
                if rt[0]:
                    console.print(":clap: do successfully")
                else:
                    console.print(f":diving_mask: failed to run, reason:{rt[1]}")
                    sys.exit(1)
            return rt

        return _wrapper

    @staticmethod
    def comparison(r1: RenderableType, r2: RenderableType) -> Table:
        table = Table(show_header=False, pad_edge=False, box=None, expand=True)
        table.add_column("1")
        table.add_column("2")
        table.add_row(r1, r2)
        return table

    @staticmethod
    def pretty_status(status: str) -> t.Tuple[str, str, str]:
        status = status.lower()
        style = "blue"
        icon = ":tractor:"
        if status == "success":
            style = "green"
            icon = ":clap:"
        elif status in ("fail", "failed"):
            style = "red"
            icon = ":fearful:"
        return status, style, icon

    @staticmethod
    def prepare_build_bundle(
        workdir: str, project: str, yaml_name: str, typ: str
    ) -> URI:
        console.print(f":construction: start to build {typ} bundle...")
        _project_uri = URI(project, expected_type=URIType.PROJECT)
        _path = os.path.join(workdir, yaml_name)
        _config = yaml.safe_load(open(_path, "r"))
        if "name" not in _config:
            raise FileFormatError(f"{_path}, no name field")

        _uri = URI.capsulate_uri(
            instance=_project_uri.instance,
            project=_project_uri.project,
            obj_type=typ,
            obj_name=_config["name"],
        )
        console.print(f":construction_worker: uri:{_uri}")
        return _uri

    @staticmethod
    def _print_history(
        title: str,
        history: t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]],
        fullname: bool = False,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        custom_header = {0: {"justify": "left", "style": "cyan", "no_wrap": True}}
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": BaseTermView.place_holder_for_empty(),
        }
        data = BaseTermView.get_history_data(history, fullname)
        BaseTermView.print_table(
            title, data, custom_header, custom_column=custom_column
        )
        return history

    @staticmethod
    def _print_info(_info: t.Dict[str, t.Any], fullname: bool = False) -> None:
        if not _info:
            console.print(":tea: not found info")
            return

        _history = _info.pop("history", (list(), dict()))

        console.rule("[green bold]Inspect Details")
        console.print(Pretty(_info, expand_all=True))

        if _history:
            console.rule("[green bold] Version History")
            BaseTermView._print_history(
                title="History List", history=_history, fullname=fullname
            )

    @staticmethod
    def _print_list(
        _bundles: t.Dict[str, t.Any], show_removed: bool = False, fullname: bool = False
    ) -> None:
        table = Table(title="Bundle List", box=box.SIMPLE, expand=True)

        table.add_column("Name")
        table.add_column("Version")
        table.add_column("Tags")
        table.add_column("Size")
        table.add_column("Runtime")
        table.add_column("Created")

        for _name, _versions in _bundles.items():
            for _v in _versions:
                if show_removed ^ _v["is_removed"]:
                    continue

                _version = (
                    _v["version"]
                    if fullname or show_removed
                    else _v["version"][:SHORT_VERSION_CNT]
                )
                if _v.get("id"):
                    _version = f"[{_v['id']:2}] {_version}"

                table.add_row(
                    _name,
                    _version,
                    ",".join(_v.get("tags", [])),
                    pretty_bytes(_v["size"]),
                    _v.get("runtime", "--"),
                    _v["created_at"],
                )

        console.print(table)

    @staticmethod
    def pretty_json(data: t.Any) -> None:
        print(json.dumps(data, indent=4, sort_keys=True))

    @staticmethod
    def list_data(
        _bundles: t.Dict[str, t.Any], show_removed: bool = False, fullname: bool = False
    ) -> t.List[t.Dict[str, t.Any]]:
        result = []
        for _name, _versions in _bundles.items():
            for _v in _versions:
                if show_removed ^ _v["is_removed"]:
                    continue

                _version = (
                    _v["version"]
                    if fullname or show_removed
                    else _v["version"][:SHORT_VERSION_CNT]
                )
                if _v.get("id"):
                    _version = f"[{_v['id']:2}] {_version}"

                result.append(
                    {
                        "name": _name,
                        "version": _version,
                        "tags": _v.get("tags", []),
                        "size": _v.get("size", 0),
                        "runtime": _v.get("runtime", ""),
                        "created_at": _v.get("created_at"),
                    }
                )

        return result

    @staticmethod
    def place_holder_for_empty(place_holder: str = "--") -> t.Callable:
        return lambda x: x or place_holder

    @staticmethod
    def print_table(
        title: str,
        data: t.List[t.Dict[str, t.Any]],
        custom_header: t.Dict[int, t.Dict] = None,
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = None,
        custom_row: t.Callable = None,
        custom_table: t.Dict = None,
    ):
        default_attr = {
            "title": title,
            "box": box.SIMPLE,
            "expand": True,
        }
        if custom_table:
            default_attr = {**default_attr, **custom_table}
        table = Table(**default_attr)  # type: ignore

        def init_header(row: t.List[str]) -> None:
            for idx, field in enumerate(row):
                extra = dict()
                if custom_header and idx in custom_header:
                    extra = custom_header[idx]
                table.add_column(snake_to_camel(field), **extra)

        header_inited = False
        for row in data:
            if not header_inited:
                init_header(list(row.keys()))
                header_inited = True
            rendered_row = list()
            for field, col in row.items():
                if custom_column and field in custom_column:
                    col = custom_column[field](col)
                rendered_row.append(col)

            row_ext: t.Dict[str, t.Any] = {}
            if custom_row:
                row_ext = custom_row(row) or {}
            table.add_row(*rendered_row, **row_ext)

        if table.row_count == 0:
            console.print("empty")
        console.print(table)

    @staticmethod
    def get_info_data(
        _info: t.Dict[str, t.Any], fullname: bool = False
    ) -> t.Dict[str, t.Any]:
        if not _info:
            return dict()

        result = _info
        _history = _info.pop("history", (list(), dict()))

        if _history:
            result["history"] = BaseTermView.get_history_data(_history, fullname)

        return result

    @staticmethod
    def get_history_data(
        history: t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]],
        fullname: bool = False,
    ) -> t.List[t.Dict]:
        result = list()
        for _h in history[0]:
            _version = _h["version"] if fullname else _h["version"][:SHORT_VERSION_CNT]
            if _h.get("id"):
                _version = f"[{_h['id']:2}] {_version}"

            result.append(
                {
                    "version": _version,
                    "tags": _h.get("tags", []),
                    "size": _h.get("size", 0),
                    "runtime": _h.get("runtime", ""),
                    "created_at": _h.get("created_at"),
                }
            )
        return result
